from . import ty
from .classes import CackResult, CackQuery, CackPredicate
from . import engine


def query_collapsed(query: ty.Dict) -> ty.Dict:
    cq = transform_query(query)
    cr = engine.query(cq)
    return transform_result(cr)


def transform_query(collapsed: ty.Dict) -> CackQuery:
    key, value = collapsed.items()[0]
    fields = [f for f in value["fields"] if not isinstance(f, dict)]
    relations = [f for f in value["fields"] if isinstance(f, dict)]
    return CackQuery(key=key, source=value["source"], filters=value["filters"], fields=fields, relations=[transform_query(r) for r in relations], required=value["required"])

def transform_result(
    result: ty.Union[None, CackResult, ty.List[CackResult]]
) -> ty.Dict:
    if result is None or result == []:
        return {None: None}
    elif isinstance(result, list):
        ((key, value),) = transform_result(result[0]).items()
        return {key: [value] + [transform_result(r)[key] for r in result[1:]]}
    else:
        fields = dict(result.fields)
        for relation in result.relations:
            ((key, value),) = transform_result(relation).items()
            if key is not None:
                fields[key] = value  # type: ignore
        return {result.key: fields}



from datetime import datetime
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

ModelType = Type  # TODO
Session = Any  # TODO
BaseModel = Any  # TODO


__all__ = [
    "Any",
    "Callable",
    "Union",
    "List",
    "Dict",
    "NoReturn",
    "Optional",
    "Sequence",
    "cast",
    "Iterator",
    "Iterable",
    "Pathlike",
    "Tuple",
    "Type",
    "TypeVar",
    "Generic",
    "Protocol",
    "ClassVar",
    "Path",
    "Set",
    "Generator",
    "datetime",
    "BaseModel",
    "ModuleType",
]



from dataclasses import dataclass, field, asdict
from enum import Enum
from . import ty

SQL_OPERATORS = [
    "is_null",
    "is_not_null",
    "==",
    "eq",
    "!=",
    "ne",
    ">",
    "gt",
    "<",
    "lt",
    ">=",
    "ge",
    "<=",
    "le",
    "like",
    "ilike",
    "not_ilike",
    "in",
    "not_in",
    "any",
    "not_any",
]


@dataclass
class CackPredicate:
    op: str
    field: str
    value: ty.Union[str, int, bool, float, None]

    def __call__(self, model):
        field = getattr(model, self.field)
        if self.op == "is_null":
            return field.is_(None) == self.value
        elif self.op == "is_not_null":
            return field.isnot(None) == self.value
        elif self.op == "==" or self.op == "eq":
            return field == self.value
        elif self.op == "!=" or self.op == "ne":
            return field != self.value
        elif self.op == ">" or self.op == "gt":
            return field > self.value
        elif self.op == "<" or self.op == "lt":
            return field < self.value
        elif self.op == ">=" or self.op == "ge":
            return field >= self.value
        elif self.op == "<=" or self.op == "le":
            return field <= self.value
        elif self.op == "like":
            return field.like(self.value)
        elif self.op == "ilike":
            return field.ilike(self.value)
        elif self.op == "not_ilike":
            return ~field.ilike(self.value)
        elif self.op == "in":
            return field.in_(self.value)
        elif self.op == "~in":
            return ~field.in_(self.value)
        elif self.op == "any":
            return field.any(self.value)
        elif self.op == "not_any":
            return ~field.any(self.value)
        else:
            raise ValueError(f"Unknown op: {self.op}")


@dataclass
class CackQuery:
    """
    Attributes
    ----------

    source (str): The name of the table (for the top-most query) or field (for relations queries) to query.
    required (bool): Whether at least one result matching the query is expected. If True and no results are found
        for the top-most query, an error is raised. If True and the query is a relation, the parent of the relation
        will be removed from the result set. If False, the return value will depend on value of `return_list`.
        If `return_list` is True, an empty list will be returned. Otherwise, the return value will be `None`.
    fields (Optional[List[str]]):
        The fields to include in the result. If None, a server-defined default view will be returned.
    key (str):
        An identifier for the query. Will be attached to the result. Defaults to "".
    filters (Dict[str, Union[str, int, bool, float, None]]): Restrict results by equality checks. Defaults to {}.
    predicates (List[CackqlPredicate]): Optional list of more complex predicates to restrict results. Defaults to [].
    return_list (Optional[bool]):
        Whether to return the results as a list. Defaults to None, which means the behaviour is determined by the server,
        based on whether the attribute has a list value or not. If multiple results are
        present and `return_list` is `False`, the first result in the list is chosen.
    slice_from (Optional[int]):
        Start index for slicing results. Slicing is performed on the result set after ordering and before
        lists are converted to scalars (if `return_list` is False). Defaults to None, which has the same semantics as 0.
    slice_to (Optional[int]): End index for slicing results. Slicing is performed on the result set after ordering and before lists are
        converted to scalars (if `return_list` is False). Defaults to None, indicating no limit.
    slice_step (Optional[int]): Step for calculating the slice. Defaults to None, which has the same semantics as 1.

    Example
    -------

    [
        {
            "key": "it's a user!"
            "source": "User",
            "fields": ["id"],
            "filters": {},
            "relations": [
                {"key": "oh, a person!", "source": "person", "filter": {}, "fields": ["name", "email"]},
                {"key": "the org", "source": "org", "filter": {}, "fields": ["name"], "relations": []}
            ]
        }
    ]
    """

    source: str
    required: bool
    fields: ty.Optional[ty.List[str]] = None
    filters: ty.Dict[str, ty.Any] = field(default_factory=dict)  # type: ignore
    predicates: ty.List[CackPredicate] = field(default_factory=list)  # type: ignore
    relations: ty.List["CackQuery"] = field(default_factory=list)  # type: ignore
    key: str = ""
    return_list: ty.Optional[bool] = None
    slice_start: ty.Optional[int] = None
    slice_stop: ty.Optional[int] = None
    slice_step: ty.Optional[int] = None

    def dict(self):
        return asdict(self)


@dataclass
class CackResult:
    key: str
    fields: ty.Dict[str, ty.Any]
    relations: ty.List[ty.Union["CackResult", ty.List["CackResult"], None]] = field(
        default_factory=list
    )

    def dict(self):
        return asdict(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['caql']

package_data = \
{'': ['*']}

install_requires = \
['pkginfo>=1.7,<2.0']

setup_kwargs = {
    'name': 'caql',
    'version': '0.5.0',
    'description': 'Tell your front-end team they have a perfectly good graph query language at home',
    'long_description': None,
    'author': 'ExplosionAI GmbH',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<3.10',
}


setup(**setup_kwargs)



from . import ty
from .classes import CackQuery, CackResult

try:
    import sqlalchemy.orm
except ImportError:  # pragma: no cover
    sqlalchemy = None


def query(
    db: ty.Session,
    cq: CackQuery,
    models: ty.Dict[str, ty.ModelType],
    authorize: ty.Callable[[ty.Any], bool],
) -> ty.Union[CackResult, ty.List[CackResult], None]:
    if sqlalchemy is not None and isinstance(db, sqlalchemy.orm.Session):
        return query_sqlalchemy(db, cq, models, authorize)
    else:
        raise NotImplementedError("Only sqlalchemy currently implemented")


def query_sqlalchemy(
    db: ty.Session,
    cq: CackQuery,
    models: ty.Dict[str, ty.ModelType],
    authorize: ty.Callable[[ty.Any], bool],
) -> ty.Union[CackResult, ty.List[CackResult], None]:
    table = models[cq.source]
    query = db.query(table).filter(*[p(table) for p in cq.predicates])
    objects = query.all()
    results = []
    for obj in objects:
        maybe_result = _maybe_make_result(cq, obj, authorize)
        if maybe_result is not None:
            results.append(maybe_result)
    results = _slice_results(results, cq)
    if cq.required and not results:
        raise ValueError("TODO message")
    if cq.return_list:
        return results
    elif results:
        return results[0]
    else:
        return None


_T = ty.TypeVar("_T")


def _slice_results(results: ty.List[_T], cq: CackQuery) -> ty.List[_T]:
    return results[slice(cq.slice_start, cq.slice_stop, cq.slice_step)]


def _maybe_make_result(
    cq: CackQuery, obj: ty.BaseModel, authorize: ty.Callable[[ty.BaseModel], bool]
) -> ty.Optional[CackResult]:
    if not authorize(obj):
        return None
    elif not _matches_filters(obj, cq.filters):
        return None
    relations = []
    for relation in cq.relations:
        maybe_result = _maybe_make_relation_results(relation, obj, authorize)
        if relation.required and not maybe_result:
            return None
        else:
            relations.append(maybe_result)
    return CackResult(
        key=cq.key, fields={k: getattr(obj, k) for k in cq.fields}, relations=relations
    )


def _maybe_make_relation_results(
    relation: CackQuery, obj: ty.BaseModel, authorize: ty.Callable[[ty.BaseModel], bool]
) -> ty.Union[CackResult, ty.List[CackResult], None]:
    if relation.predicates:
        raise NotImplementedError("Predicates currently only available on outer query")
    attr = getattr(obj, relation.source)
    values = attr if isinstance(attr, list) else [attr]
    values = [
        v for v in values if authorize(v) and _matches_filters(v, relation.filters)
    ]
    values = _slice_results(values, relation)
    maybe_results = [_maybe_make_result(relation, v, authorize) for v in values]
    results = [r for r in maybe_results if r is not None]
    if _get_return_list(relation.return_list, attr):
        return results
    elif results:
        return results[0]
    else:
        return None


def _get_return_list(
    return_list: ty.Optional[bool], attr: ty.Union[ty.List, ty.BaseModel]
) -> bool:
    """Handle smart default for return_list: if it's None, check whether the attribute
    is a list.
    """
    if return_list is not None:
        return return_list
    elif isinstance(attr, list):
        return True
    else:
        return False


def _matches_filters(
    obj: ty.BaseModel,
    filters: ty.Dict[str, ty.Union[str, int]],
) -> bool:
    return all(getattr(obj, key) == value for key, value in filters.items())
